                PAGE	60,132		;
                NAME    CLOCK           ;
                TITLE   CLOCK - Sperry PC-1 and PC/HT Clock Driver ;
;  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *
;  *                                                                 *
;  *           THE FOLLOWING PROGRAM IS THE SOLE PROPERTY OF         *
;  *                         RONALD Q. SMITH                         *
;  *        CONTAINING HIS PROPRIETARY CONFIDENTIAL INFORMATION      *
;  *                  COPYRIGHT RONALD Q. SMITH 1984                 *
;  *                                                                 *
;  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *
		COMMENT !

PROGRAM:CLOCK

CLOCK is a clock driver for use with the Sperry PC-1 and PC/HT clock.  IT also
supports the Leading Edge Model D and other Mitsubishi equivalents.  It provides
direct support for a clock functions through the TIME and DATE DOS commands.
No other programs are required to read and set the clock.

The driver is configured in CONFIG.SYS with the statement:
DEVICE=CLOCK.SYS

At system initialization, the clock driver will display the current time and
date.  At that time you will not be asked to change the values.  You may,
however, put DATE and TIME commands in an AUTOEXEC.BAT file.  This will give you
an opportunity to change the values.
!
		SUBTTL	Constants and Data	;
		PAGE			;
;
; Data structure definitions.
;
; FAR pointer
;
FAR_DAT         STRUC                   ;
OFF             DW      ?               ; Offset
SG              DW      ?               ; Segment
FAR_DAT         ENDS                    ;
;
; DOS interface constants.
;
DOS$            EQU     21H             ; DOS function call
PRINT$          EQU     9               ; Print string function
;
; BIOS interface constants.
;
TOD$            EQU     1AH             ; BIOS clock interrupt
GET_TIME$       EQU     2               ; Read time function
SET_TIME$       EQU     3               ; Write time function
GET_DATE$       EQU     4               ; Read date function
SET_DATE$       EQU     5               ; Write date function
BUSY_CLK        EQU     0FFH            ; Busy status
;
; DOS call to CLOCK parameter packet format.
;
PARM_SG         SEGMENT AT 0            ; Dummy segment definition
REQ_LNG         DB      ?               ; Total length of request packet
UNIT            DB      ?               ; Relative unit number this request
CMD_COD         DB      ?               ; Command code
STATUS          DW      ?               ; Return status
ERR             EQU     8000H           ; Error status-check low order byte
DONE            EQU     80H             ; Request finished
WRP             EQU     0               ; Write protect violation
UNT_ERR         EQU     1               ; Illegal unit number
NOT_RDY         EQU     2               ; Device not ready
CMD_ERR         EQU     3               ; Illegal command code
CRC_ERR         EQU     4               ; CRC error-parity error
PKT_ERR         EQU     5               ; Illegal packet length
SK_ERR          EQU     6               ; Seek error
MED_ERR         EQU     7               ; Illegal media format byte
SEC_ERR         EQU     8               ; Sector not found
WRT_ERR         EQU     10              ; General write error
RD_ERR          EQU     11              ; General read error
GEN_ERR         EQU     12              ; General device error
                DB      8 DUP(?)        ; Reserved for DOS
PKT             EQU     $-REQ_LNG       ; Start of variable fields
UNIT_CNT        DB      ?               ; INIT-number of units
END_LOC         FAR_DAT <>              ; INIT-last address used by driver
                ORG     PKT             ;
MED_DESC        DB      ?               ; Media descriptor(not used)
XFER_LOC        FAR_DAT <>              ; INPUT,WRITE-buffer loc
BYTE_CNT        DW      ?               ; INPUT,WRITE-Number of bytes
                                        ; to read or write,read or written
PARM_SG         ENDS                    ;
;
; Format of parameter packet for clock drivers.
;
DATA            SEGMENT AT 0            ; Dummy segment
DAYS            DW      ?               ; Days since 1-1-80
MIN             DB      ?               ; Minute of hour
HRS             DB      ?               ; Hour of day
HSEC            DB      ?               ; Hundredths of second
SEC             DB      ?               ; Second of minute
DATA            ENDS                    ;
;
; Special characters.
;
CR_CHR          EQU     13              ; Carriage return
LF_CHR          EQU     10              ; Line feed
;
; Start of driver segment.
;
CLOCK           SEGMENT PARA PUBLIC     ;
;
; Device driver header for DOS.
; Must be at beginning of segment.
;
                DD      -1              ; Last driver in file
                DW      8008H           ; Not block, CLOCK$ driver
                DW      STRAT           ; Strategy entry point
                DW      INTRPT          ; Interrupt entry point
                DB      "CLOCK$  "      ; Name of driver
PROG_NAME       DB      "CLOCK    "     ;
                DB      "Copyright 1991 - Ronald Q. Smith" ;
;
; Command code vector.
;
FUN_TAB         LABEL   WORD            ;
                DW      INIT            ; Initialization
                DW      FUN_BAD         ; Not used
                DW      FUN_BAD         ; Not used
                DW      FUN_BAD         ; Not used
                DW      READ            ; Read data
                DW      FUN_BAD         ;
                DW      FUN_BAD         ;
                DW      FUN_BAD         ;
                DW      WRITE           ; Write data
NFUN            EQU     ($-FUN_TAB)/2   ; Highest legal command code
;
; Minimum legal packet size for command type.
;
PKT_SIZ         DB      22              ; INIT
                DB      15              ; Not used
                DB      22              ; Not used
                DB      22              ; Not used
                DB      22              ; Read
                DB      14              ; Not used
                DB      13              ; Not used
                DB      13              ; Not used
                DB      22              ; Write
;
; Pointer save area.
;
PARM            FAR_DAT <>              ; Packet pointer saved by STRAT
;
; Constants for computation.
;
C365            DW      365             ; Days in a year
C7              DW      7               ; Days in a week
MON_TAB         DW      0,31,59,90,120,151,181,212,243,273,304,334 ; Days in
                                        ; year prior to this month
                SUBTTL  Code for CLOCKDRV ;
                PAGE                    ;
;
; Start of code for CLOCKDRV.
;
		ASSUME CS:CLOCK,DS:NOTHING,ES:NOTHING ;
;
;
; PROCEDURE: STRAT
;
; STRAT saves the parameter packet pointer for the next invocation of
; INTRPT.  This two step operation is a DOS standard even if not very useful
; for current device handlers, especially those that use the BIOS.  For
; interrupt driven devices, this permits queuing an operation immediately.
;
; INPUT:
;       ES:BX   =       Parameter packet address
;
; OUTPUT:
;	ALL registers unchanged.
;
STRAT           PROC    FAR             ;
;
; Copy packet address to save area.
;
                MOV     PARM.OFF,BX     ; Offset
                MOV     PARM.SG,ES      ; Segment
                RET                     ;
STRAT           ENDP                    ;
;
;
; PROCEDURE: INTRPT
;
; INTRPT is called by DOS to perform the function.  The function must be
; complete when INTRPT returns.  The function to be performed is defined
; by the packet given on the last STRAT call.
;
; DOS provides a stack large enough to save all registers, but no larger.
;
; INPUT:
;       From last STRAT call
;
; OUTPUT:
;       Through packet.
;       All registers unchanged.
;
INTRPT          PROC    FAR             ;
;
; Save registers used.
;
                CLD                     ;
                PUSH    AX              ;
                PUSH    BX              ;
                PUSH    CX              ;
                PUSH    DX              ;
                PUSH    DI              ;
                PUSH    SI              ;
                PUSH    BP              ;
                PUSH    DS              ;
                PUSH    ES              ;
;
; Set up segment registers.
;
                MOV     AX,CS           ;
                MOV     DS,AX           ; Data in same segment
                ASSUME  DS:CLOCK        ;
;
; Set up parameter pointer.
;
                LES     BX,DWORD PTR PARM ;
                ASSUME  ES:PARM_SG      ;
;
; Check command code and packet length.
;
                MOV     AL,CMD_COD[BX]  ; Command code from packet
                CMP     AL,NFUN         ; Within range?
                JAE     CMD_BAD         ; No
                CBW                     ;
                MOV     SI,AX           ;
                MOV     AL,REQ_LNG[BX]  ; Length this packet
                CMP     AL,PKT_SIZ[SI]  ; Long enough?
                JB      PKT_BAD         ; No
                SHL     SI,1            ; Word index
                CALL    FUN_TAB[SI]     ; Call function procedure
;
; Function completed.
; Set completion status, restore registers and return.
;
SET_DONE:                               ;
                OR      STATUS[BX],DONE ; Function completed
                POP     ES              ;
                POP     DS              ;
                POP     BP              ;
                POP     SI              ;
                POP     DI              ;
                POP     DX              ;
                POP     CX              ;
                POP     BX              ;
                POP     AX              ;
                RET                     ;
;
; Local errors.
;
CMD_BAD:                                ; Illegal command code
                OR      STATUS[BX],ERR+CMD_ERR ; 
                JMP     SET_DONE        ;
PKT_BAD:                                ; Illegal packet length
                OR      STATUS[BX],ERR+PKT_ERR ;
                JMP     SET_DONE        ;
INTRPT          ENDP                    ;
;
;
; PROCEDURE: FUN_BAD
;
; FUN_BAD is called for an illegal command code error which is detected 
; by the function table.
;
; INPUT:
;       None
;
; OUTPUT:
;       STATUS set to command error
;
FUN_BAD         PROC                    ;
                POP     AX              ; Throw away return
                JMP     CMD_BAD         ; Go return error status
FUN_BAD         ENDP                    ;
;
;
; PROCEDURE: READ
;
; READ is called to read the clock of the Sperry PC-1 and PC/HT, and return the
; correct date and time.
;
; INPUT:
;       BYTE_CNT =      Number of bytes to do
;       XFER_LOC =      Buffer address to do I/O from/into
;
; OUTPUT:
;       BYTE_CNT =      6
;       XFER_LOC =      Current date and time
;
READ            PROC                    ;
                MOV     DI,2            ; Retry counter
                CMP     BYTE_CNT[BX],6  ; Is requested byte count enough
                JAE     RD_LOOP         ; Yes
CLK_RDERR:                              ;
                MOV     STATUS[BX],ERR+RD_ERR ; Set general read error
                RET                     ;
;
; Start over if we get a busy status as the time has changed.
;
RD_LOOP:                                ;
                DEC     DI              ; Done all retries?
                JS      CLK_RDERR       ; Yes, error
                MOV     AH,GET_DATE$    ; Read date
                INT     TOD$            ;
                CMP     AL,BUSY_CLK     ; Clock busy?
                JE      RD_LOOP         ;
;
; Convert the date to the form DOS requires (days since 1980-1-1)
; CL = Year (mod 1980)
; DH = Month
; DL = Day
;
                MOV     AL,DH           ; Month
                SHL     AL,1            ;
                CBW                     ;
                MOV     SI,AX           ; Month * 2 index
                MOV     CH,DL           ; Save days
                MOV     AL,CL           ;
                CBW                     ; Date to days since 1980-1-1
                MUL     C365            ; Years * 365
                ADD     AX,MON_TAB[SI-2] ; + days to start of this month
                ADD     AL,CH           ; + days
                ADC     AH,0            ;
                TEST    CL,3            ; Leap year?
                JNZ     NOT_LEAP        ; No
                CMP     SI,4            ; Past leap day?
                JA      NOT_LEAP        ; Yes
                DEC     AX              ; Remove extra day
NOT_LEAP:                               ;
                SHR     CL,1            ; Number of leap years past
                SHR     CL,1            ;
                ADD     AL,CL           ;
                ADC     AH,0            ;
                MOV     SI,AX           ; Save days
                MOV     AH,GET_TIME$    ; Get time
                INT     TOD$            ;
                CMP     AL,BUSY_CLK     ; Busy?
                JE      RD_LOOP         ;
                MOV     DL,50           ; 0.50 Seconds since we don't know
                OR      AL,AL           ; 24-hour clock?
                JZ      NOT_PM          ; Yes
                CMP     CH,12           ; 12 O'clock?
                JNE     NOT_12          ; No
                XOR     CH,CH           ; Reset to 0
NOT_12:                                 ;
                CMP     AL,3            ; PM?
                JNE     NOT_PM          ;
                ADD     CH,12           ; Convert to 24-hour clock
NOT_PM:                                 ;
                MOV     AX,SI           ;
                LDS     SI,DWORD PTR XFER_LOC[BX] ;
                ASSUME  DS:DATA         ;
                MOV     DAYS[SI],AX     ;
                MOV     WORD PTR MIN[SI],CX ; Minutes and hours
                MOV     WORD PTR HSEC[SI],DX ; Seconds and 1/100ths
                PUSH    CS              ; Restore data segment
                POP     DS              ;
                ASSUME  DS:CLOCK        ;
                MOV     BYTE_CNT[BX],6  ; Put number of bytes moved
                RET                     ; Read done
READ            ENDP                    ;
;
;
; PROCEDURE: WRITE
;
; WRITE gets the date and time from the user and sets the clock.
;
; INPUT:
;       XFER_LOC =      Location of user's data
;       BYTE_CNT =      Number of bytes from user(must be 6)
;
; OUTPUT:
;       Updated time and date
;
                ASSUME  DS:CLOCK        ;
WRITE           PROC                    ;
;
; Check that data length is correct.
;
                MOV     DI,4            ; Retry counter
                CMP     BYTE_CNT[BX],6  ; Must be 6 bytes
                JAE     DO_WRITE        ; OK
CLK_WRERR:                              ;
                OR      STATUS[BX],ERR+WRT_ERR ; Return error status
                RET                     ;
DO_WRITE:                               ;
                DEC     DI              ; More retries?
                JS      CLK_WRERR       ; No, error
;
; Convert DOS data to format clock requires.
;
                LDS     SI,DWORD PTR XFER_LOC[BX] ; Location of user buffer
                ASSUME  DS:DATA         ;
                MOV     DX,WORD PTR HSEC[SI] ; 1/100ths second, seconds
                MOV     CX,WORD PTR MIN[SI] ; minutes, hours
                MOV     SI,DAYS[SI]     ; days since 1980-1-1
                PUSH    CS              ;
                POP     DS              ;
                ASSUME  DS:CLOCK        ;
                ADD     DX,256-50       ; Round seconds up
                CMP     DH,60           ;
                JB      NO_ROUND        ;
                XOR     DH,DH           ;
                INC     CL              ;
                CMP     CL,60           ;
                JB      NO_ROUND        ;
                XOR     CL,CL           ;
                INC     CH              ;
                CMP     CH,24           ;
                JB      NO_ROUND        ;
                XOR     CH,CH           ;
                INC     SI              ;
NO_ROUND:                               ;
                MOV     AH,SET_TIME$    ;
                XOR     AL,AL           ; 24-hour clock
                INT     TOD$            ;
                CMP     AL,BUSY_CLK     ; Clock busy?
                JE      DO_WRITE        ; Yes, try again
WRITE_DATE:                             ;
                DEC     DI              ;
                JS      CLK_WRERR       ;
                MOV     AX,SI           ; Convert days to yr, mon, day, dow
                INC     AX              ; Because 1980-1-1 was Tue
                INC     AX              ;
                XOR     DX,DX           ;
                DIV     C7              ; Get day of week
                MOV     CH,DL           ;
                MOV     AX,SI           ;
                XOR     DX,DX           ;
                DIV     C365            ; Year = Days / 365
                MOV     CL,AL           ;
                DEC     AX              ; Don't count this leap year
                JS      YR_OK           ;
                SHR     AL,1            ;
                SHR     AL,1            ; Number of leap years past
                SUB     DX,AX           ; Remove leap days
                JG      YR_OK           ; End of a year?
                DEC     CL              ; One year less
                ADD     DX,C365         ; Add back the year's days
YR_OK:                                  ;
                MOV     SI,2            ; Month index
                TEST    CL,3            ; Is this a leap year?
                JNZ     MON_LP          ; No
                CMP     DX,60           ; Past the leap day?
                JB      MON_LP          ; No
                MOV     SI,4            ;
                JE      MON_FND         ; If Feb 29
                DEC     DX              ; Remove leap day
MON_LP:                                 ;
                CMP     DX,MON_TAB[SI]  ; Is it this month?
                JBE     MON_FND         ; Yes
                INC     SI              ; Next month
                INC     SI              ;
                CMP     SI,24           ; Done?
                JB      MON_LP          ; No
MON_FND:                                ;
                SUB     DX,MON_TAB[SI-2] ; Get day of month
                MOV     AX,SI           ;
                SHR     AL,1            ;
                MOV     DH,AL           ; Month
                MOV     AH,CL           ; Get leap year indicator
                AND     AH,3            ; 0 = leap year
                MOV     AL,4            ; 1 = year 3 of cycle
                SUB     AL,AH           ; 2 = year 2 of cycle
                AND     AL,3            ; 3 = year 1 of cycle
                MOV     AH,SET_DATE$    ;
                INT     TOD$            ; Set date
                CMP     AL,BUSY_CLK     ;
                JE      WRITE_DATE      ; If busy, try again
                MOV     BYTE_CNT[BX],6  ; 6 bytes actually written
                RET                     ; Write operation done
WRITE           ENDP                    ;
;
;
; PROCEDURE: INIT
;
; INIT is called once at the beginning of execution(boot time) by DOS.
; It must do any necessary internal initialization and return information
; to DOS on the last address used by the driver.  INIT also does the 
; initialization and release to DOS of any extra memory above the DIP
; switch setting.  INIT will print the current date and time.
; Since this code is not needed after the first time, the last address returned
; is set before the INIT procedure.  Constants that are not or may not be
; needed are placed here so that they may also be released.
;
; INPUT:
;       NONE
;
; OUTPUT:
;       END_LOC  =      First address not needed
;
INIT_ST         EQU     $               ;
;
; Initialization message.  It can also be released.
;
INIT_MSG        DB      "DATE/TIME = "  ;
CENT_MSG        DB      "19"            ;
YR_MSG          DB      "80 "           ;
MON_MSG         DB      "Jan "          ;
DAY_MSG         DB      "01 "           ;
DOW             DB      "Sun "          ;
HR_MSG          DB      "00"            ;
MIN_MSG         DB      "00:"           ;
SEC_MSG         DB      "00"            ;
                DB      CR_CHR,LF_CHR,"$" ;
STK             DB      128 DUP ("S")   ; Local stack segment
STK_PTR         EQU     $               ;
SAV_STK         FAR_DAT <>              ;
MON_BYT         LABEL   BYTE            ; Character month abbreviations
                DB      "Jan"           ;
                DB      "Feb"           ;
                DB      "Mar"           ;
                DB      "Apr"           ;
                DB      "May"           ;
                DB      "Jun"           ;
                DB      "Jul"           ;
                DB      "Aug"           ;
                DB      "Sep"           ;
                DB      "Oct"           ;
                DB      "Nov"           ;
                DB      "Dec"           ;
DOW_BYT         LABEL   BYTE            ; Character day of week abbreviations
                DB      "Sun"           ;
                DB      "Mon"           ;
                DB      "Tue"           ;
                DB      "Wed"           ;
                DB      "Thu"           ;
                DB      "Fri"           ;
                DB      "Sat"           ;
ERR_MSG         DB      "Unable to access calendar clock." ;
                DB      CR_CHR,LF_CHR,"$" ;
C100            DB      100             ;
;
; Initialization procedure.
;
INIT            PROC                    ;
;
; Set up local stack for DOS calls.
;
                MOV     SAV_STK.SG,SS   ; Save DOS stack
                MOV     SAV_STK.OFF,SP  ;
                MOV     AX,CS           ;
                MOV     SS,AX           ; Set up local stack
                MOV     SP,OFFSET STK_PTR ;
;
; Get current time and date and edit into message.
;
                MOV     DI,4            ; Retry counter
INT_LOOP:                               ;
                DEC     DI              ;
                JNS     TRY_AGAIN       ;
                JMP     BAD_CLK         ;
TRY_AGAIN:                              ;
                MOV     AH,GET_DATE$    ;
                INT     TOD$            ;
                CMP     AL,BUSY_CLK     ; Clock busy?
                JE      INT_LOOP        ; Then try again
                MOV     AL,DH           ;
                CBW                     ;
                MOV     SI,AX           ; Month * 3
                SHL     AX,1            ;
                ADD     SI,AX           ;
                MOV     AX,WORD PTR MON_BYT[SI-3] ;
                MOV     WORD PTR MON_MSG,AX ;
                MOV     AL,MON_BYT[SI-1] ; Put ASCII month
                MOV     MON_MSG+2,AL    ;
                MOV     AL,CH           ; Day of week * 3
                CBW                     ;
                MOV     SI,AX           ;
                SHL     AX,1            ;
                ADD     SI,AX           ;
                MOV     AX,WORD PTR DOW_BYT[SI] ;
                MOV     WORD PTR DOW,AX ;
                MOV     AL,DOW_BYT[SI+2] ;
                MOV     DOW+2,AL        ; Put ASCII day of week
                MOV     AL,DL           ; Day
                CALL    BTD             ;
                MOV     WORD PTR DAY_MSG,AX ;
                MOV     AL,CL           ; Year
                CBW                     ;
                ADD     AX,1980         ;
                DIV     C100            ; Separate century
                MOV     CH,AH           ;
                CALL    BTD             ;
                MOV     WORD PTR CENT_MSG,AX ;
                MOV     AL,CH           ;
                CALL    BTD             ;
                MOV     WORD PTR YR_MSG,AX ;
                MOV     AH,GET_TIME$    ;
                INT     TOD$            ;
                CMP     AL,BUSY_CLK     ; Busy?
                JE      INT_LOOP        ; Yes, try again
                TEST    AL,2            ; 12-hour clock?
                JZ      NOT_12HR        ; No
                CMP     CH,12           ; Is it 12?
                JNE     NOT_12A         ; No
                XOR     CH,CH           ;
NOT_12A:                                ;
                CMP     AL,3            ; Is it PM?
                JNE     NOT_12HR        ; No
                ADD     CH,12           ;
NOT_12HR:                               ;
                MOV     AL,CH           ;
                CALL    BTD             ; Hours
                MOV     WORD PTR HR_MSG,AX ;
                MOV     AL,CL           ; Minutes
                CALL    BTD             ;
                MOV     WORD PTR MIN_MSG,AX ;
                MOV     AL,DH           ; Seconds
                CALL    BTD             ;
                MOV     WORD PTR SEC_MSG,AX ;
;
; Print initialization message and current date and time.
;
                MOV     DX,OFFSET INIT_MSG ;
INIT_DONE:                              ;
                MOV     AH,PRINT$       ; Print string function code
                INT     DOS$            ; Print the message
;
; Return first location not needed.
;
                MOV     END_LOC.OFF[BX],OFFSET INIT_ST ; Start of init area
                MOV     END_LOC.SG[BX],CS ; Segment
;
; Restore DOS stack.
;
                MOV     SS,SAV_STK.SG   ;
                MOV     SP,SAV_STK.OFF  ;
                RET                     ;
BAD_CLK:                                ;
                MOV     DX,OFFSET ERR_MSG ; Display error message instead
                JMP     INIT_DONE       ; of date and time
INIT            ENDP                    ;  
;
;
; PROCEDURE: BTD
;
; Converts binary numbers between 0 and 99 to ASCII decimal.
;
; INPUT:
;       AL      =       binary number
;
; OUTPUT:
;       AX      =       ASCII decimal number(units in AH)
;
BTD             PROC                    ;
                AAM                     ; Convert to Decimal
                OR      AX,"00"         ; Convert to ASCII
                XCHG    AL,AH           ; Get in order for string storage
                RET                     ;
BTD             ENDP                    ;
CLOCK		ENDS			;
;  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *
;  *                                                                 *
;  *           THE PRECEDING PROGRAM IS THE SOLE PROPERTY OF         *
;  *                         RONALD Q. SMITH                         *
;  *        CONTAINING HIS PROPRIETARY CONFIDENTIAL INFORMATION      *
;  *                  COPYRIGHT RONALD Q. SMITH 1984                 *
;  *                                                                 *
;  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *
		END			;
